const sequelize = require('sequelize');
const Notifications = require('../models/notifications');
const Users = require('../models/users');
const secureFilter = require('../utils/secureFilter');
const config = require('../config');

const formatWhere = where => (
    Object.keys(where).map(field => `${field}=${JSON.stringify(where[field])}`).join(' and ')
);

const get = async (ctx, next) => {
    ctx.assert(ctx.state.user, 403, 'User not logged in!');
    ctx.assert(await ctx.state.checkPermission('user_notifications', 'READ'), 403, 'User not allowed to read notifications.');
    try {
        ctx.status = 200;
        ctx.body.unReadCount = await Notifications.count({
            where: {userid: ctx.state.user.id, readtime: null, dismissed: null}
        });

        const where = {userid: ctx.state.user.id, dismissed: null};

        if (ctx.query.created_after) {
            where.createdat = {
                [sequelize.Op.gt]: new Date(ctx.query.created_after)
            };
        }

        const notifications = await Notifications.findAll({
            where,
            attributes: Notifications.fields.filter(field => field.exposed).map(f => f.name),
            order: [
                ['createdat', 'DESC']
            ],
        });
        ctx.body.items = notifications
            .map(notification => notification.get({plain: true}))
            .map(notification => Object.assign(
                {},
                notification,
                {
                    fields: notification.fields || {}
                }
            ));

        // Only store the initial get notification not subsequent data refreshes.
        if (ctx.state.auditLog && !ctx.query.created_after && config.ENABLE_VERBOSE_AUDIT) {
            ctx.state.auditLog('User retrieved notifications.');
        }
    } catch (err) {
        ctx.status = 500;
        ctx.body.error = err.message;
    }
    return next();
};

const filtersList = async (ctx, next) => {
    ctx.assert(ctx.state.user, 403, `User not logged in! ${JSON.stringify(ctx.state)}`);
    try {
        ctx.status = 200;
        const where = (
            ctx.request.query.where ?
                secureFilter(ctx.request.body.where, Users.fields.filter(field => field.exposed)) :
                {}
        );
        const environments = await Users.findAll({
            attributes: ['environment', [sequelize.fn('count', sequelize.col('environment')), 'count']],
            group: ['environment'],
            where: {environment: {[sequelize.Op.ne]: null}}
        });
        const organizations = await Users.findAll({
            attributes: ['orgid', 'environment', [sequelize.fn('count', sequelize.col('orgid')), 'count']],
            group: ['orgid', 'environment'],
            where: Object.assign({}, where, {orgid: {[sequelize.Op.ne]: null}})
        });
        const personas = await Users.findAll({
            attributes: ['persona', [sequelize.fn('count', sequelize.col('persona')), 'count']],
            group: ['persona'],
            where: Object.assign({}, where, {persona: {[sequelize.Op.ne]: null}})
        });
        const user_roles = await Users.findAll({
            attributes: ['user_role', [sequelize.fn('count', sequelize.col('user_role')), 'count']],
            group: ['user_role'],
            where: Object.assign({}, where, {user_role: {[sequelize.Op.ne]: null}})
        });
        ctx.body.filters = [
            {
                name: 'environment',
                options: environments.map(env => ({
                    name: env.get('environment'),
                    value: env.get('environment'),
                    reach: parseInt(env.get('count'), 10)
                }))
            },
            {
                name: 'orgid',
                dependsOn: 'environment',
                multiple: true,
                options: organizations.map(org => ({
                    name: String(org.get('orgid')),
                    value: org.get('orgid'),
                    environment: org.get('environment'),
                    reach: parseInt(org.get('count'), 10)
                }))
            },
            personas && personas.length > 0 && {
                name: 'persona',
                options: personas.map(persona => ({
                    name: persona.get('persona'),
                    value: persona.get('persona'),
                    reach: parseInt(persona.get('count'), 10)
                }))
            },
            user_roles && user_roles.length > 0 && {
                name: 'user_role',
                options: user_roles.map(user_role => ({
                    name: user_role.get('user_role'),
                    value: user_role.get('user_role'),
                    reach: parseInt(user_role.get('count'), 10)
                }))
            }
        ].filter(f => f);
        if (ctx.state.auditLog) {
            ctx.state.auditLog('User retrieved permissible filters.');
        }
    } catch (err) {
        ctx.status = 500;
        ctx.body.error = err.message;
    }
    return next();
};

const reach = async (ctx, next) => {
    ctx.assert(ctx.state.user, 403, 'User not logged in!');
    ctx.assert(await ctx.state.checkPermission('csr.notifications', 'CREATE'), 403, 'User not allowed to create notifications.');
    try {
        const where = (
            ctx.request.body.where ?
                secureFilter(ctx.request.body.where, Users.fields.filter(field => field.exposed)) :
                {}
        );
        ctx.body.count = await Users.count({where});
        if (ctx.state.auditLog) {
            ctx.state.auditLog(`User retrieved a count of users reached with filter: ${JSON.stringify(where)}.`);
        }
        ctx.status = 200;
    } catch (err) {
        ctx.body.error = err.message;
    }
    return next();
};

const post = async (ctx, next) => {
    ctx.assert(ctx.state.user, 403, `User not logged in! ${JSON.stringify(ctx.state)}`);
    ctx.assert(await ctx.state.checkPermission('csr.notifications', 'CREATE'), 403, 'User not allowed to create notifications.');
    try {
        const type = ctx.request.body.type || 'announcement';
        if (!ctx.request.body.title) {
            throw new Error('Please provide a title.');
        }
        if (!ctx.request.body.content) {
            throw new Error('Please provide content.');
        }
        if (
            ctx.request.body.action &&
            ctx.request.body.action.url &&
            !ctx.request.body.action.url.match(/^https?:\/\/([a-z0-9-]+\.)*(?:carbonblack|vmware|community.broadcom|broadcom)\.com(?:\/.*)?$/i) &&
            !ctx.request.body.action.url.match(/^\/(.+)/)
        ) {
            throw new Error('Only Cb Domains can currently be linked to.');
        }
        const attributes = ['id', 'subscriptions'];
        const where = (
            ctx.request.body.where ?
                secureFilter(ctx.request.body.where, Users.fields.filter(field => field.exposed)) :
                {}
        );

        let offset = 0;
        let limit = config.MAX_NOTIFICATION_CREATION_CHUNK_SIZE;
        let total = Number.MAX_VALUE;

        while (offset < total) {
            const {count, rows: targetUsers} = await Users.findAndCountAll({where, attributes, offset, limit});
            total = count;
            offset += limit;

            const action = JSON.stringify(ctx.request.body.action);
            const fields = JSON.stringify(ctx.request.body.fields);

            if (targetUsers.length > 0) {
                const notifications = [];
                targetUsers.forEach((user) => {
                    const subscriptions = Object.assign({}, Users.defaultSubscriptions, user.get('subscriptions'), Users.overideSubscriptions);
                    const deliveries = (
                        subscriptions[type] &&
                        subscriptions[type].enabled &&
                        subscriptions[type].delivery
                    ) || [];

                    if (deliveries.includes('in_app')) {
                        notifications.push({
                            type,
                            action,
                            fields,
                            title: ctx.request.body.title,
                            content: ctx.request.body.content,
                            userid: user.get('id'),
                            createdat: new Date(),
                            createdby: ctx.state.user.id
                        });
                    }
                });
                let success = true;
                // Async processing
                Notifications.bulkCreate(notifications).then(() => {
                    console.log(`Created ${notifications.length} notifications in bulk!`);
                }).catch((error) => {
                    console.error('Failed to insert notifications!', error.message, ctx.request);
                    success = false;
                });
                if (success) {
                    if (ctx.state.auditLog) {
                        ctx.state.auditLog(`User posted ${notifications.length} in-product notification(s).`);
                    }
                    Notifications.create({
                        type: 'job_success',
                        title: 'Announcement sent',
                        content: `${notifications.length} notifications sent ("${ctx.request.body.title}" - to users where ${formatWhere(where)})`,
                        userid: ctx.state.user.id,
                        createdat: new Date(),
                        createdby: ctx.state.user.id
                    });
                } else {
                    if (ctx.state.auditLog) {
                        ctx.state.auditLog('User attempted posting in-product notifications with failure.');
                    }
                    Notifications.create({
                        type: 'job_failed',
                        title: 'Failed to send announcement',
                        content: `Failed to send notification "${ctx.request.body.title}". Error: ${error.message}`,
                        userid: ctx.state.user.id,
                        createdat: new Date(),
                        createdby: ctx.state.user.id
                    });
                }
                ctx.body.message = 'Operation queued.';
            } else {
                ctx.body.message = 'No users matching criteria.';
            }
            ctx.status = 200;
        }
    } catch (err) {
        ctx.status = err.message === 'Conflict' ? 409 : 400;
        ctx.body.error = err.message;
    }

    return next();
};

const patch = async (ctx, next) => {
    ctx.assert(ctx.state.user, 403, `User not logged in! ${JSON.stringify(ctx.state)}`);
    ctx.assert(await ctx.state.checkPermission('user_notifications', 'READ'), 403, 'User not allowed to read notifications.');
    try {
        const filter = {where: {userid: ctx.state.user.id, id: ctx.params.id}};
        const editableFields = (
            Notifications.fields
                .filter(f => f.editable)
                .map(f => f.name)
        );
        const update = Object.keys(ctx.request.body)
            .filter(key => editableFields.includes(key))
            .reduce((obj, key) => {
                obj[key] = ctx.request.body[key];
                return obj;
            }, {});
        ctx.body.update = await Notifications.update(update, filter);

        if (ctx.state.auditLog && config.ENABLE_VERBOSE_AUDIT) {
            ctx.state.auditLog('User updated a notification.');
        }
        ctx.status = 200;
    } catch (err) {
        ctx.status = 400;
        ctx.body.error = err.message;
    }
    return next();
};

module.exports = {
    '/reach': {
        post: {
            description: 'Returns a count of users that will receive the notification.',
            run: reach,
            responses: {
                200: {
                    content: {
                        schema: {
                            type: 'object',
                            properties: {
                                count: {
                                    type: 'number',
                                    description: 'The total number of users reached.'
                                }
                            }
                        }
                    }
                }
            }
        }
    },
    // Temporary hack as the UI route was invontarily updated
    '/notifications/filters': {
        get: {
            description: 'Returns a list of permissible filters to narrow the reach of a notification, along with associated reach for each filter.',
            run: filtersList,
            responses: {
                200: {
                    content: {
                        schema: {
                            type: 'object',
                            properties: {
                                filters: {
                                    type: 'array',
                                    description: 'A list of permissible filters including options and reach.'
                                }
                            }
                        }
                    }
                }
            }
        }
    },
    '/filters': {
        get: {
            description: 'Returns a list of permissible filters to narrow the reach of a notification, along with associated reach for each filter.',
            run: filtersList,
            responses: {
                200: {
                    content: {
                        schema: {
                            type: 'object',
                            properties: {
                                filters: {
                                    type: 'array',
                                    description: 'A list of permissible filters including options and reach.'
                                }
                            }
                        }
                    }
                }
            }
        }
    },
    '/notifications': {
        post: {
            description: 'Post a new notification',
            run: post
        },
        get: {
            description: 'Returns a list of notification',
            run: get
        }
    },
    '/notifications/:id': {
        patch: {
            description: 'Update an existing notification',
            run: patch
        }
    }
};
