const sequelize = require('sequelize');
const Notifications = require('../models/notifications');
const Users = require('../models/users');
const config = require('../config');

const summary = async (ctx, next) => {
    if (ctx.query.key !== config.STATS_KEY) {
        return ctx.throw(403);
    }
    try {
        ctx.status = 200;
        ctx.body.totalUsers = await Users.count();
        ctx.body.totalNotifications = await Notifications.count();
        ctx.body.totalUnreadNotifications = await Notifications.count({
            where: {
                readtime: null
            }
        });
        ctx.body.totalDismissedNotifications = await Notifications.count({
            where: {
                dismissed: true
            }
        });
        const environments = await Users.findAll({
            attributes: ['environment', [sequelize.fn('count', sequelize.col('environment')), 'count']],
            group: ['environment'],
        });
        ctx.body.environments = {};
        await Promise.all(environments.map(async (env) => {
            const envStats = {};
            const users = await Users.findAll({ where: { environment: env.environment } });
            envStats.totalUsers = users.length;
            envStats.totalNotifications = await Notifications.count({ where: {
                userid: {
                    [sequelize.Op.in]: users.map(user => user.id)
                }
            } });
            envStats.totalUnreadNotifications = await Notifications.count({ where: {
                readtime: null,
                userid: {
                    [sequelize.Op.in]: users.map(user => user.id)
                }
            } });
            envStats.totalDismissedNotifications = await Notifications.count({ where: {
                dismissed: true,
                userid: {
                    [sequelize.Op.in]: users.map(user => user.id)
                }
            } });
            ctx.body.environments[env.environment] = envStats;
        }));
        const latestNotifications = await Notifications.findAll({
            attributes: ['title', [sequelize.fn('count', sequelize.col('title')), 'sent']],
            group: ['title']
            // limit: 20
        });
        ctx.body.topNotifications = await Promise.all(
            latestNotifications
                .map(notification => ({
                    title: notification.get('title'),
                    sent: parseInt(notification.get('sent'), 10)
                }))
                .sort((a, b) => {
                    if (a.sent > b.sent) {
                        return -1;
                    } else if (a.sent === b.sent) {
                        return 0;
                    }
                    return 1;
                })
                .slice(0, 50)
                .map(
                    async notification => ({
                        title: notification.title,
                        sent: notification.sent,
                        unread: await Notifications.count({
                            where: {
                                title: notification.title,
                                readtime: null
                            }
                        }),
                        dismissed: await Notifications.count({
                            where: {
                                title: notification.title,
                                dismissed: true
                            }
                        })
                    })
                )
        );
        if (ctx.state.auditLog) {
            ctx.state.auditLog('User accessed the metrics about notification service.');
        }
    } catch (err) {
        ctx.status = 500;
        ctx.body.error = err.message;
    }
    return next();
};

module.exports = {
    '/stats/summary': {
        get: {
            description: 'Returns a some useful metrics about the service.',
            run: summary,
            responses: {
                200: {
                    content: {
                        schema: {
                            type: 'object',
                            properties: {
                                totalUsers: {
                                    type: 'number',
                                    description: 'The total number of users.'
                                },
                                totalNotifications: {
                                    type: 'number',
                                    description: 'The total number of notifications.'
                                },
                                totalUnreadNotifications: {
                                    type: 'number',
                                    description: 'The total number of unread notifications.'
                                },
                                totalDismissedNotifications: {
                                    type: 'number',
                                    description: 'The total number of dismissed notifications.'
                                },
                                environment: {
                                    type: 'object',
                                    description: 'Notifications per environments and other environment metrics.'
                                },
                                topNotifications: {
                                    type: 'object',
                                    description: 'List of most read notifications and respective stats.'
                                }
                            }
                        }
                    }
                }
            }
        }
    }
};
