const Consumer = require('sqs-consumer');
const AWS = require('aws-sdk');
const sequelize = require('sequelize');
const { v4: uuidv4 } = require('uuid');
const config = require('../config');
const Notifications = require('../models/notifications');
const NotificationHistory = require('../models/notificationHistory');
const Users = require('../models/users');
const staticHostsMap = require('../utils/staticHostsMap');
const alertLink = require('../utils/alertLink');
const { formatEmail, sendEmail } = require('../utils/emailGateway');
const https = require('https');
const saveNotificationHistory = require('../utils/notificationHistory');
const CB_ANALYTICS = 'CB Analytics';
const WATCHLIST = 'Watchlist';


if (!config.AWS_DEFAULT_REGION) {
    throw new Error('Missing AWS REGION.');
}

AWS.config.update({ region: config.AWS_DEFAULT_REGION });
const service = {
    workers: {},
    sqs: {},
    messagesProcessed: 0,
    queues: {}
};

/**
 * Lookup map for ANALYST_STATE (determination) to the proper presentation values
 */
const ANALYST_STATE_PRESENTATION_LOOKUP = {
    "TRUE_POSITIVE": "Likely threat",
    "FALSE_POSITIVE": "Unlikely threat"
};

/**
 * Lookup map for ANALYST_TRIAGE_INFO (workflow state) to the proper presentation values
 */
const ANALYST_TRIAGE_INFO_PRESENTATION_LOOKUP = {
    "UNCLAIMED": "Unclaimed",
    "IN_PROGRESS": "In progress",
    "TRIAGE_COMPLETE": "Triage complete",
    "ACTION_REQUESTED": "Action requested",
    "PENDING_RESPONSE": "Pending MDR response",
    "RECEIVED_RESPONSE": "Received MDR response",
    // Deprecated value (should only see RECEIVED_RESPONSE)
    "RESPONSE_RECEIVED": "Received MDR response"
};

const ALERT_NOTIFICATION_TYPES = ['POLICY_ACTION', 'TTP', 'THREAT_SCORE', 'WATCHLIST', 'MDR'];
const USER_DELIVERY_OPTIONS = ['EMAIL', 'IN_APP'];
const CONNECTOR_DELIVERY_OPTIONS = ['SIEM', 'PRIVATE_INTELLIGENCE', 'THREAT_EXCHANGE', 'CRITS'];
const ONE_DAY = 24 * 60 * 60 * 1000;

/**
 * env name to region
 * if not listed explicitly, default is used.
 */
const ENV_TO_AWS_REGION = {
    default: 'us-east-1',
    prod06: 'eu-central-1',
    prodnrt: 'ap-northeast-1',
    prodsyd: 'ap-southeast-2'
};

const floodFilterOrgKeyList = new Set(config.EMAIL_FLOOD_FILTER_ORG_KEY_LIST.trim().split(","));

/**
 * Looks up the region for an env
 *
 * @param env name (eg prod02)
 * @returns {string}
 */
const getRegionForEnv = env => {
    if (ENV_TO_AWS_REGION[env]) {
        return ENV_TO_AWS_REGION[env];
    } else {
        // use aws default region set as environment variable
        return config.AWS_DEFAULT_REGION;
    }
}

/**
 * Look up the url of a queue based on its name
 *
 * @param region region where the queue is defined
 * @param queueName name of the queue
 */
const getQueueUrl = async (region, queueName) => {
    if (config.SQS_QUEUE_URL) {
        console.log(`Using static queue url ${config.SQS_QUEUE_URL}`)
        return config.SQS_QUEUE_URL
    }

    if (queueName.includes('localdev')) {
        return `${config.LOCALDEV_SQS_ENDPOINT}/queue/localdev_notification_queue`;
    }
    if (service.queues[queueName]) {
        console.log(`Getting queue url for Queue=${queueName} in Region=${region} Url=${service.queues[queueName]}`)
        return service.queues[queueName];
    }

    console.log(`Getting queue url for Queue=${queueName} in Region=${region} via getQueueUrl`)
    const req = service.sqs[region].getQueueUrl({ QueueName: queueName });
    const response = await req.promise();
    service.queues[queueName] = response.QueueUrl;
    return service.queues[queueName];
};

const sendMailAndCreateHistory = async (recipient, html, text, subject, historyItem, type) => {
    try {
        await sendEmail({
            to: recipient.address,
            html,
            text,
            subject
        });

        // Add date local var to ensure date consistency between log and history
        const currentDate = new Date();

        // Enable logs in non-gov environments
        if (!config.GOV && config.EMAIL_SHOW_LOGS) {
            console.log(`Email sent to: ${recipient.address}, subject: ${subject}, ruleId: ${historyItem.alertRuleId}, orgKey: ${historyItem.orgKey},  ruleType: ${type}, time: ${currentDate.toLocaleString()}`);
        }

        historyItem.deliveryEmail = recipient.address;
        historyItem.deliveryEmailSent = currentDate;
    } catch (err) {
        historyItem.deliveryEmailStatus = err.message;
        console.error(`Email send error: ${err.message}`);
    }
}

// Look up the dashboard url for a given environment
// This is used as the static resources root for emails
// (think: images in the email are relative to this url)
const getDashboardUrl = (env) => {
    if (config.DASHBOARD_URL) {
        return config.DASHBOARD_URL
    }

    return staticHostsMap[env];
}

const getTitle = (type, context) => {
    if (type === 'MDR') {
        if(context.hasOwnProperty('THREAT_HUNT_LIKELY_THREAT_FOUND')) {
            // For both hasMTHEntitlement == true and hasMTHEntitlement == false we should use following
            if(context.THREAT_HUNT_LIKELY_THREAT_FOUND === true)
                return 'Managed Threat Hunting - Likely Threats Found';
            else
                return 'Managed Threat Hunting - No Threats Found';
        }
        else if(context.hasOwnProperty('ANALYST_COMMENT')) {
            if(hasMDREntitlement(context) && (isCbAnalyticsAlert(context) || isWatchlistAlert(context)))
                return 'Managed Detection & Response - Comment Added';
            else if(hasMDEntitlement(context) && isCbAnalyticsAlert(context))
                return 'Managed Detection - Comment Added';
            else if(hasMTHEntitlement(context) && isWatchlistAlert(context))
                return 'Managed Threat Hunting - Comment Added';
            else if(isCbAnalyticsAlert(context))
                return 'Managed Detection & Response - Comment Added';
            else if(isWatchlistAlert(context))
                return 'Managed Threat Hunting - Comment Added';
            else
                return 'Managed Detection & Response - Comment Added';
        }
        else if(context.hasOwnProperty('ANALYST_STATE') && context.ANALYST_STATE === 'TRUE_POSITIVE') {
            if(hasMDREntitlement(context) && (isCbAnalyticsAlert(context) || isWatchlistAlert(context)))
                return 'Managed Detection & Response - Likely Threat';
            else if(hasMDEntitlement(context) && isCbAnalyticsAlert(context))
                return 'Managed Detection - Likely Threat';
            else if(hasMTHEntitlement(context) && isWatchlistAlert(context))
                return 'Managed Threat Hunting - Likely Threat';
            else if(isCbAnalyticsAlert(context))
                return 'Managed Detection & Response - Likely Threat';
            else if(isWatchlistAlert(context))
                return 'Managed Threat Hunting - Likely Threat';
            else
                return 'Managed Detection & Response - Likely Threat';
        }
        else if (context.hasOwnProperty('ANALYST_STATE') && context.ANALYST_STATE === 'FALSE_POSITIVE'){
            if(hasMDREntitlement(context) && (isCbAnalyticsAlert(context) || isWatchlistAlert(context)))
                return 'Managed Detection & Response - Unlikely Threat';
            else if(hasMDEntitlement(context) && isCbAnalyticsAlert(context))
                return 'Managed Detection - Unlikely Threat';
            else if(hasMTHEntitlement(context) && isWatchlistAlert(context))
                return 'Managed Threat Hunting - Unlikely Threat';
            else if(isCbAnalyticsAlert(context))
                return 'Managed Detection & Response - Unlikely Threat';
            else if(isWatchlistAlert(context))
                return 'Managed Threat Hunting - Unlikely Threat';
            else
                return 'Managed Detection & Response - Unlikely Threat';
        }
    }

    return type === 'POLICY_ACTION' ? 'Policy Action Taken' : 'Carbon Black Cloud Alert';
}

const formatAndSendEmail = async(templateName, context, recipient, historyItem) => {
    const {html, text, subject} = formatEmail(templateName, context);
    await sendMailAndCreateHistory(recipient, html, text, subject, historyItem, context.type);
}

const hasMDREntitlement = (context) => {
    return context.hasOwnProperty("HAS_MDR_ENTITLEMENT") && context.HAS_MDR_ENTITLEMENT === true;
}

const hasMDEntitlement = (context) => {
    return context.hasOwnProperty("HAS_MD_ENTITLEMENT") && context.HAS_MD_ENTITLEMENT === true;
}

const hasMTHEntitlement = (context) => {
    return context.hasOwnProperty("HAS_MTH_ENTITLEMENT") && context.HAS_MTH_ENTITLEMENT === true;
}

const isCbAnalyticsAlert = (context) => {
    return context.hasOwnProperty("MDR_ALERT_TYPE") && context.MDR_ALERT_TYPE === 'CB Analytics';
}

const isWatchlistAlert = (context) => {
    return context.hasOwnProperty("MDR_ALERT_TYPE") && context.MDR_ALERT_TYPE === 'Watchlist';
}

const getLinkLabel = (type, context) => {
    if (type === 'MDR') {
        if (context.hasOwnProperty('ANALYST_COMMENT')){
            return 'Log in to View Comment';
        }
        else {
            return 'Log in to View Alert';
        }
    }

    return 'View in Carbon Black Cloud';
}

const formatType = type => type.split('_').map(w => w[0].toUpperCase() + w.slice(1)).join(' ');

const handleMessage = (environment, region) => async (message) => {
    try {
        // decode message.body to utf-8 string
        const decoded = message.Body && JSON.parse(Buffer.from(message.Body, "base64").toString("utf-8"));
        if (!decoded.recepients || decoded.recepients.length === 0) {
            console.log('Warning: Message with no recipient', decoded.context.THREAT_ID);
            return;
        }

        if (config.DEBUG) {
            console.log('Processing message:', JSON.stringify(decoded));
        }

        // Inject THREAT_SCORE alerts
        if (ALERT_NOTIFICATION_TYPES.includes(decoded.type) && decoded.recepients) {
            const recipients = decoded.recepients.filter(r => (
                USER_DELIVERY_OPTIONS.includes(r.type)
            ));
            const connectors = decoded.recepients.filter(r => (
                CONNECTOR_DELIVERY_OPTIONS.includes(r.type)
            ));
            const emailRecipients = decoded.recepients.filter(r => (
                r.type === 'EMAIL'
            ));
            const userDetails = await Users.findAll({ where: {
                orgid: decoded.orgId,
                environment
            } });
            const subscribers = recipients.map((recipient) => {
                const user = userDetails.find(u => (
                    (recipient.loginId && u.loginid === recipient.loginId)
                ));

                return Object.assign(
                    {},
                    recipient,
                    user && user.toJSON()
                );
            });
            if (subscribers && subscribers.length > 0) {
                // Insert the notifications
                const notifications = [];
                const notificationHistoryItems = [];
                await Promise.all(subscribers.map(async (recipient) => {
                    const context = Object.assign(
                        {},
                        decoded.context,
                        {
                            environment,
                            type: decoded.type,
                            staticHost: getDashboardUrl(environment),
                            title: getTitle(decoded.type, decoded.context)
                        }
                    );
                    const historyItem = {
                        sha256Hash: context.SHA256_HASH,
                        eventTime: context.EVENT_TIME,
                        eventId: uuidv4(),
                        deviceId: context.DEVICE_ID,
                        alertRuleId: context.RULE_ID,
                        incidentId: context.INCIDENT_ID,
                        orgId: decoded.orgId,
                        orgKey: decoded.orgKey,
                        alertRuleName: context.RULE_NAME,
                        threatId: context.THREAT_ID,
                        loginId: recipient.loginId,
                        correlationId: decoded.correlation_id,
                        comment: decoded.comment,
                        emails: decoded.emails,
                        connectorId: decoded.connector_id,
                        status: "SENT"
                    };

                    // https://jira.carbonblack.local/browse/DSER-15410
                    // Support passing explicit links from CBD so the
                    // notification service doesn't have to guess
                    // or show multiple links when it is unable to figure it out.
                    //
                    // It is still this service's responsibility however to add the hostname.
                    // Note: Future work is to possibly do localization
                    // here in this service, also to allow more than one explicit link.
                    if (context.EXPLICIT_LINKS && context.PRIMARY_LINK_KEY) {
                        context.links = [
                            {
                                url: `${getDashboardUrl(environment)}${context.EXPLICIT_LINKS[context.PRIMARY_LINK_KEY]}`,
                                label: getLinkLabel(decoded.type, decoded.context)
                            }
                        ];
                    }
                    else if(decoded.type === 'MDR' && context.hasOwnProperty("THREAT_HUNT_LIKELY_THREAT_FOUND")){
                        const investigateLinkLabel = 'Log In to View Hits';
                        const alertLinkLabel = 'Log In to View Likely Threat Alerts';

                        context.threatHuntInvestigatePageLinks = [
                            {
                                url: `${getDashboardUrl(environment)}${context.THREAT_HUNT_INVESTIGATE_PAGE_LINK}`,
                                label: investigateLinkLabel
                            }
                        ];

                        if(context.THREAT_HUNT_LIKELY_THREAT_FOUND === true){
                            context.threatHuntAlertsPageLinks = [
                                {
                                    url: `${getDashboardUrl(environment)}${context.THREAT_HUNT_ALERT_PAGE_LINK}`,
                                    label: alertLinkLabel
                                }
                            ];
                        }
                    }


                    // Generate Alert Link based on host, alert details.
                    if (context.type === 'POLICY_ACTION') {
                        const qs = [
                            's[searchWindow]=ALL'
                        ];
                        if (context.EVENT_ID) {
                            qs.push(`s[c][QUERY_STRING_TYPE][0]=${context.EVENT_ID}`);
                        }
                        if (context.DEVICE_ID) {
                            qs.push(`s[c][DEVICE_ID][0]=${context.DEVICE_ID}`);
                        }
                        if (!context.links) {
                            context.links = [
                                {
                                    url: `${context.staticHost}/investigate?${qs.join('&')}`,
                                    label: 'My company uses VMware Carbon Black Cloud Endpoint Standard'
                                },
                                {
                                    // alert_id:${context.EVENT_ID}%20and%
                                    url: `${context.staticHost}/threat-hunter/investigate?query=device_id:${context.DEVICE_ID}`,
                                    label: 'My company uses VMware Carbon Black Cloud Endpoint Standard + Enterprise EDR'
                                }
                            ];
                            context.linksLabel = 'View in console';
                        }
                    } else if (context.type === 'THREAT_SCORE' || context.type === 'TTP') {
                        if (!context.links) {
                            context.links = [
                                {
                                    url: `${context.staticHost}/investigate?s[searchWindow]=ALL&s[c][DEVICE_ID][0]=${context.DEVICE_ID}&s[c][INCIDENT_ID][0]=${context.INCIDENT_ID}`,
                                    label: 'View in Carbon Black Cloud'
                                }
                            ];
                        }
                    }
                    else if (!context.links) {
                        context.links = [
                            {
                                url: alertLink(context),
                                label: 'View in Carbon Black Cloud'
                            }
                        ];
                    }

                    // Populate the human-readable values for MDR workflow/determination
                    if (context.hasOwnProperty('ANALYST_STATE')) {
                        // ANALYST_STATE is for determination
                        context.ANALYST_STATE_PRESENTATION = ANALYST_STATE_PRESENTATION_LOOKUP[context.ANALYST_STATE];
                    }
                    if (context.hasOwnProperty('ANALYST_TRIAGE_INFO')) {
                        // ANALYST_TRIAGE_INFO is for workflow
                        context.ANALYST_TRIAGE_INFO_PRESENTATION = ANALYST_TRIAGE_INFO_PRESENTATION_LOOKUP[context.ANALYST_TRIAGE_INFO];
                    }

                    if (recipient.type === 'IN_APP') {
                        if (!recipient.id) {
                            historyItem.deliveryInAppStatus = 'Could not map notification to user.';
                            console.error(`IN_APP notification type with no user id found! ${JSON.stringify({ recipient, decoded })}`);
                        } else {
                            historyItem.deliveryInAppStatus = 'Queued for delivery';
                            notifications.push({
                                type: decoded.type.toLowerCase(),
                                title: `${formatType(decoded.type)}${decoded.context.INCIDENT_ID ? ` - ${decoded.context.INCIDENT_ID}` : ''}`,
                                content: decoded.context.INCIDENT_SUMMARY,
                                action: JSON.stringify(context.links[0]),
                                fields: JSON.stringify(decoded.context),
                                userid: recipient.id,
                                createdat: new Date()
                            });
                        }
                    } else if (recipient.type === 'EMAIL') {
                        const orgKey = decoded.orgKey;
                        if (config.EMAIL_FLOOD_FILTER_ENABLED && orgKey !== null && floodFilterOrgKeyList.size > 0 && floodFilterOrgKeyList.has(orgKey)) {
                           console.log(`Email flood setting enabled for the orgKey ${orgKey}, Skipping the email`)
                        } else if (!config.ENABLE_EMAIL_SUPPORT) {
                            historyItem.deliveryEmailStatus = 'Email support disabled.';
                            console.warn('EMAIL notification but ENABLE_EMAIL_SUPPORT was disabled (this is OK - message will be forwarded to legacy service)');
                        } else if (!recipient.address) {
                            historyItem.deliveryEmailStatus = 'No email address specified.';
                            console.error('EMAIL notification with no specified address!', recipient, JSON.stringify(decoded));
                        } else {
                            const totalEmailsReceived = config.ENABLE_RATE_LIMITING && (
                                await NotificationHistory.count({ where: {
                                    delivery_email: recipient.address,
                                    createdat: {
                                        [sequelize.Op.gt]: new Date(Date.now() - ONE_DAY)
                                    }
                                } })
                            );

                            if (
                                config.ENABLE_RATE_LIMITING &&
                                totalEmailsReceived >= config.MAX_EMAILS_PER_DAY
                            ) {
                                historyItem.deliveryEmailStatus = 'Max emails per day limit reached.';
                                const filter = { where: {
                                    delivery_email: recipient.address,
                                    deliveryEmailStatus: 'Max emails per day limit reached.',
                                    createdat: {
                                        [sequelize.Op.gt]: new Date(Date.now() - ONE_DAY)
                                    }
                                } };
                                const maxEmailsWarningSent = await NotificationHistory
                                    .count(filter);
                                if (maxEmailsWarningSent === 0) {

                                    const { html, text, subject } = formatEmail('emailQuotaExceeded', {
                                        MAX_EMAILS_PER_DAY: config.MAX_EMAILS_PER_DAY
                                    });

                                    console.warn(`Max emails per day reached for ${recipient.address}`);

                                    await sendEmail({
                                        to: recipient.address,
                                        html,
                                        text,
                                        subject
                                    });
                                }
                            }
                            else {
                                if(decoded.type === 'MDR'){
                                    if(context.hasOwnProperty("THREAT_HUNT_LIKELY_THREAT_FOUND")){
                                        // For both hasMTHEntitlement == true and hasMTHEntitlement == false we should use following
                                        if(context.THREAT_HUNT_LIKELY_THREAT_FOUND === true){
                                            await formatAndSendEmail('mdrThreatHuntThreatFound', context, recipient, historyItem);
                                        }
                                        else{
                                            await formatAndSendEmail('mdrThreatHuntNoThreatFound', context, recipient, historyItem);
                                        }
                                    }
                                    else if(context.hasOwnProperty('ANALYST_COMMENT')) {
                                        if(hasMDREntitlement(context) && isCbAnalyticsAlert(context)) {
                                            await formatAndSendEmail('mdrCommentAddedForCbAnalytics', context, recipient, historyItem);
                                        } else if(hasMDREntitlement(context) && isWatchlistAlert(context)) {
                                            await formatAndSendEmail('mdrCommentAddedForWatchlist', context, recipient, historyItem);
                                        } else if(hasMDEntitlement(context) && isCbAnalyticsAlert(context)){
                                            await formatAndSendEmail('mdCommentAddedForCbAnalytics', context, recipient, historyItem);
                                        } else if(hasMTHEntitlement(context) && isWatchlistAlert(context)) {
                                            await formatAndSendEmail('mthCommentAddedForWatchlist', context, recipient, historyItem);
                                        } else if(isCbAnalyticsAlert(context)) {
                                            await formatAndSendEmail('mdrCommentAddedForCbAnalytics', context, recipient, historyItem);
                                        } else if(isWatchlistAlert(context)) {
                                            await formatAndSendEmail('mthCommentAddedForWatchlist', context, recipient, historyItem);
                                        } else
                                            await formatAndSendEmail('mdrCommentAdded', context, recipient, historyItem);
                                    }
                                    else if(context.hasOwnProperty('ANALYST_STATE') && context.ANALYST_STATE === 'TRUE_POSITIVE') {
                                        if(hasMDREntitlement(context) && isCbAnalyticsAlert(context)) {
                                            await formatAndSendEmail('mdrLikelyThreatForCbAnalytics', context, recipient, historyItem);
                                        } else if(hasMDREntitlement(context) && isWatchlistAlert(context)){
                                            await formatAndSendEmail('mdrLikelyThreatForWatchlist', context, recipient, historyItem);
                                        } else if(hasMDEntitlement(context) && isCbAnalyticsAlert(context)){
                                            await formatAndSendEmail('mdLikelyThreatForCbAnalytics', context, recipient, historyItem);
                                        } else if(hasMTHEntitlement(context) && isWatchlistAlert(context)) {
                                            await formatAndSendEmail('mthLikelyThreatForWatchlist', context, recipient, historyItem);
                                        } else if(isCbAnalyticsAlert(context)) {
                                            await formatAndSendEmail('mdrLikelyThreatForCbAnalytics', context, recipient, historyItem);
                                        } else if(isWatchlistAlert(context)) {
                                            await formatAndSendEmail('mthLikelyThreatForWatchlist', context, recipient, historyItem);
                                        } else
                                            await formatAndSendEmail('mdrLikelyThreat', context, recipient, historyItem);
                                    }
                                    else if (context.hasOwnProperty('ANALYST_STATE') && context.ANALYST_STATE === 'FALSE_POSITIVE'){
                                        if(hasMDREntitlement(context) && isCbAnalyticsAlert(context)) {
                                            await formatAndSendEmail('mdrUnlikelyThreatForCbAnalytics', context, recipient, historyItem);
                                        } else if(hasMDREntitlement(context) && isWatchlistAlert(context)){
                                            await formatAndSendEmail('mdrUnlikelyThreatForWatchlist', context, recipient, historyItem);
                                        } else if(hasMDEntitlement(context) && isCbAnalyticsAlert(context)){
                                            await formatAndSendEmail('mdUnlikelyThreatForCbAnalytics', context, recipient, historyItem);
                                        } else if(hasMTHEntitlement(context) && isWatchlistAlert(context)) {
                                            await formatAndSendEmail('mthUnlikelyThreatForWatchlist', context, recipient, historyItem);
                                        } else if(isCbAnalyticsAlert(context)) {
                                            await formatAndSendEmail('mdrUnlikelyThreatForCbAnalytics', context, recipient, historyItem);
                                        } else if(isWatchlistAlert(context)) {
                                            await formatAndSendEmail('mthUnlikelyThreatForWatchlist', context, recipient, historyItem);
                                        } else
                                            await formatAndSendEmail('mdrUnlikelyThreat', context, recipient, historyItem);
                                    }
                                }
                                else if(decoded.type === 'POLICY_ACTION'){
                                    await formatAndSendEmail('policyActionApplied', context, recipient, historyItem);
                                }
                                else if(context.IS_MDR_ELIGIBLE_ALERT === true){
                                    if(context.MDR_ALERT_TYPE === 'Watchlist')
                                        await formatAndSendEmail('pscMDREligibleWatchlistAlert', context, recipient, historyItem);
                                    else if(context.MDR_ALERT_TYPE === 'CB Analytics')
                                        await formatAndSendEmail('pscMDREligibleCBAnalyticsAlert', context, recipient, historyItem);
                                }
                                else{
                                    await formatAndSendEmail('pscAlert', context, recipient, historyItem);
                                }
                                if (config.DEBUG) {
                                    console.log(`Send email to ${recipient.address}`);
                                }
                            }
                        }
                    } else {
                        console.log('Unknown recipient type', recipient);
                    }
                    if (config.ENABLE_NOTIFICATION_HISTORY) {
                        notificationHistoryItems.push(historyItem);
                    }
                }));
                if (notifications.length > 0) {
                    try {
                        await Notifications.bulkCreate(notifications);
                        notificationHistoryItems
                            .filter(item => item.deliveryInAppStatus === 'Queued for delivery')
                            .forEach((item, index) => {
                                notificationHistoryItems[index].deliveryInAppStatus = 'Delivered';
                                notificationHistoryItems[index].deliveryInAppSent = Date.now();
                            });
                    } catch (err) {
                        console.error('Failed to insert in-product notifications!', err.message);
                        notificationHistoryItems
                            .filter(item => item.deliveryInAppStatus === 'Queued for delivery')
                            .forEach((item, index) => {
                                notificationHistoryItems[index].deliveryInAppStatus = 'Insertion failure';
                                notificationHistoryItems[index].comment = err.message;
                            });
                    }
                }
                if (config.ENABLE_NOTIFICATION_HISTORY && notificationHistoryItems.length > 0) {
                    try {
                        await saveNotificationHistory({'data': notificationHistoryItems})
                    } catch (err) {
                        console.error(`Failed writing to notification history! ${err.message}`);
                    }
                }
            }
        } else {
            // We don't yet know how to handle this message type
            console.warn(`Unhandled message type: ${decoded.type}`);
        }
        service.messagesProcessed++;
    } catch (e) {
        // An unknown error occurred, proceed anyway with old service (could be permission or other)
        console.error(`Processing error: ${e.message}`, JSON.stringify(message));
    }
};
const createWorker = async (environment, number) => {
    const region = getRegionForEnv(environment);
    const sqs = service.sqs[region];
    const queueName = config.SQS_QUEUE_NAME ? config.SQS_QUEUE_NAME : `${environment}_notification_service_queue`;

    try {
        const QueueUrl = await getQueueUrl(region, queueName);
        console.log(`Got queue url ${QueueUrl}`)

        const consume = handleMessage(environment, region);
        const worker = Consumer.create({
            batchSize: config.QUEUE_BATCH_SIZE,
            queueUrl: QueueUrl,
            attributeNames: [
                'All'
            ],
            handleMessage: consume,
            // Processes the batch messages in parallel
            handleMessageBatch: messages => Promise.all(messages.map(consume)),
            sqs
        });
        worker.on('error', (err) => {
            console.error(`Worker ${environment}:${number}:${QueueUrl} ERROR: "${err.message}"`);
        });
        // If first worker for this environment, log out SQS stats.
        if (number === 0) {
            setInterval(() => {
                sqs.getQueueAttributes({
                    QueueUrl,
                    AttributeNames: [
                        'ApproximateNumberOfMessages',
                        'ApproximateNumberOfMessagesNotVisible',
                        'ApproximateNumberOfMessagesDelayed'
                    ]
                }, (err, data) => {
                    if (err) {
                        console.error(`${environment}:${number} ${QueueUrl} ERROR: "${err.message}"`);
                    } else if (
                            data.Attributes.ApproximateNumberOfMessagesNotVisible > 0 ||
                            data.Attributes.ApproximateNumberOfMessages > 0
                    ) {
                        console.warn(`${environment}:${number}: ~messages: ${data.Attributes.ApproximateNumberOfMessages}, ~messages not visible: ${data.Attributes.ApproximateNumberOfMessagesNotVisible}`);
                    }
                });
            }, 10000);
        }
        return worker;
    } catch (error) {
        console.log(`Failed to initialize worker ${environment}:${number}:${region}:${queueName} ${error.message}`, error);

        return null;
    }
};
const logOutput = () => {
    const seconds = (Date.now() - service.lastBatchStart) / 1000;
    const mps = service.messagesProcessed > 0 ?
        Math.round((service.messagesProcessed / seconds) * 100) / 100 :
        0;
    const totalWorkers = Object.values(service.workers).length;

    if (mps > 0) {
        console.log(`${(new Date()).toLocaleString()} [${config.ACTIVE_ENVIRONMENTS}:${totalWorkers} workers] ${service.messagesProcessed} messages processed since last check in (${mps}mps)`);
    }
    service.lastBatchStart = Date.now();
    service.messagesProcessed = 0;
};

const init = async () => {
    const environments = config.ACTIVE_ENVIRONMENTS.split(',');
    const maxWorkers = config.MAX_QUEUE_WORKERS;
    for (let i = 0; i < environments.length; i++) {
        const environment = environments[i];
        const region = getRegionForEnv(environment);
        if (environment === 'localdev') {
            AWS.config.update({
                endpoint: config.LOCALDEV_SQS_ENDPOINT,
                httpOptions: { agent: new https.Agent({ rejectUnauthorized: false }) }
            });
        }
        if (!service.sqs[region]) {
            console.log('Created SQS client for', region);
            service.sqs[region] = new AWS.SQS({ apiVersion: '2012-11-05', signatureVersion: 'v4', region });
        }
        for (let x = 0; x < maxWorkers; x++) {
            const worker = await createWorker(environment, x);
            if (worker) {
                service.workers[`${environment}:${x}`] = worker;
            }
        }
        console.log(`${(new Date()).toLocaleString()} Notifications Queue Consumer Setup ${environment} (${config.MAX_QUEUE_WORKERS} workers)`);
    }
};

service.start = async () => {
    await init();

    const count = await Notifications.count();
    if (count > 100000000) {
        console.warn('There are more than 100m notifications in the store.');
    }
    service.lastBatchStart = Date.now();
    Object.values(service.workers).forEach((consumer) => {
        try {
            consumer.start();
        } catch (error) {
            console.error('Failed to start sqs consumer:', error.message);
        }
    });
    service._logger = setInterval(logOutput, 30000);
    console.log(`${(new Date()).toLocaleString()} Notifications Queue Consumer started with ${Object.values(service.workers).length} workers.`);
};
service.stop = async () => {
    Object.values(service.workers).forEach(consumer => consumer.stop());
    if (service._logger) {
        clearInterval(service._logger);
    }
    console.log(`${(new Date()).toLocaleString()} Notifications Queue Consumer stopped.`);
};

module.exports = service;
