const sequelize = require('sequelize');
const Notifications = require('../models/notifications');
const config = require('../config');
const chai = require('chai');
const service = require('../jobs/cleanup');

const expect = chai.expect;

const ONE_DAY = 86400000;
const readMaxAge = ONE_DAY * config.MAX_READ_NOTIFICATIONS_AGE;
const allMaxAge = ONE_DAY * config.MAX_NOTIFICATIONS_AGE;

/* eslint no-unused-expressions: 0 */
describe('Testing notification cleanup script', async () => {
    // Create notifications of various 'readtime' and 'createat' times
    const testObjects = [
        { readtime: new Date(), createdat: new Date(), content: 'New read notification' },
        { readtime: null, createdat: new Date(), content: 'New unread notification' },
        { readtime: new Date(), createdat: new Date((new Date()).getTime() - readMaxAge), content: 'Max read age' },
        { readtime: null, createdat: new Date((new Date()).getTime() - allMaxAge), content: 'Max notification age' }
    ];
    const testNotifications = [];

    before(async () => {
        testObjects.forEach((test) => {
            testNotifications.push({
                type: 'announcement',
                title: 'Test cleanup',
                content: test.content,
                readtime: test.readtime,
                createdat: test.createdat
            });
        });
        // Add test notifications to db
        await Notifications.bulkCreate(testNotifications);
        // Run cleanup script
        await service.run();
    });

    after(async () => {
        await Notifications.destroy({ truncate: true, cascade: false });
    });

    it(`Should delete read notifications after ${config.MAX_READ_NOTIFICATIONS_AGE} days`, async () => {
        const notifications = await Notifications.findAll({
            where: {
                readtime: {
                    [sequelize.Op.ne]: null
                },
                createdat: {
                    [sequelize.Op.lt]: new Date((new Date()).getTime() - readMaxAge)
                }
            }
        }).map(notification => notification.get({ plain: true }));

        const message = `Expected read notifications after ${config.MAX_READ_NOTIFICATIONS_AGE} days to be deleted.
        Actual: ${JSON.stringify(notifications, null, 4)}`;
        expect(notifications, message).to.be.empty;
    });

    it(`Should delete any notification after ${config.MAX_NOTIFICATIONS_AGE} days`, async () => {
        const notifications = await Notifications.findAll({
            where: {
                createdat: {
                    [sequelize.Op.lt]: new Date((new Date()).getTime() - allMaxAge)
                }
            }
        }).map(notification => notification.get({ plain: true }));
        const message = `Expected all notifications after ${config.MAX_NOTIFICATIONS_AGE} days to be deleted.
        Actual: ${JSON.stringify(notifications, null, 4)}`;
        expect(notifications, message).to.be.empty;
    });

    it(`Should not delete read notifications before ${config.MAX_READ_NOTIFICATIONS_AGE} days`, async () => {
        const notifications = await Notifications.findAll({
            where: {
                readtime: {
                    [sequelize.Op.ne]: null
                },
                createdat: {
                    [sequelize.Op.gt]: new Date((new Date()).getTime() - readMaxAge)
                }
            }
        }).map(notification => notification.get({ plain: true }));

        const message = `Expected no read notifications before ${config.MAX_READ_NOTIFICATIONS_AGE} days to be deleted`;
        expect(notifications, message).to.not.be.empty;
        const content = 'New read notification';
        const notification = notifications.find(item => item.content === content);
        expect(notification, `Expected notification with content '${content}' to not be deleted`).to.not.be.undefined;
    });

    it(`Should not delete unread notification before ${config.MAX_NOTIFICATIONS_AGE} days`, async () => {
        const notifications = await Notifications.findAll({
            where: {
                readtime: {
                    [sequelize.Op.eq]: null
                },
                createdat: {
                    [sequelize.Op.gt]: new Date((new Date()).getTime() - allMaxAge)
                }
            }
        }).map(notification => notification.get({ plain: true }));

        const message = `Expected no notifications before ${config.MAX_NOTIFICATIONS_AGE} days to be deleted`;
        expect(notifications, message).to.not.be.empty;
        const content = 'New unread notification';
        const notification = notifications.find(item => item.content === content);
        expect(notification, `Expected notification with content '${content}' to not be deleted`).to.not.be.undefined;
    });
});
