const chai = require('chai');
const { formatEmail } = require('../utils/emailGateway');
const context = require('../test-scripts/sample_notification');
const crAlertContext = require('../test-scripts/sample_cr_alert_notification')
const crAlertHbfwContext = require('../test-scripts/sample_cr_alert_notification_HBFW')
const crAlertIdsNtaContext = require('../test-scripts/sample_cr_alert_notification_IDS_NTA')
const crAlertMDRContext = require('../test-scripts/sample_cr_alert_notification_MDR')
const crAlertMDRCommentAdded = require('../test-scripts/sample_alert_notification_MDR_comment_added')
const crAlertMDRLikelyThreat = require('../test-scripts/sample_alert_notification_MDR_Likely_Threat')
const crAlertMDRUnlikelyThreat = require('../test-scripts/sample_alert_notification_MDR_Unlikely_Threat')
const crAlertMDRThreatHuntLikelyThreatFound = require('../test-scripts/sample_cr_alert_notification_MDR_Threat_Hunt_likely_threat_found')
const crAlertMDRThreatNoThreatsFound = require('../test-scripts/sample_cr_alert_notification_MDR_Threat_Hunt_No_threat_found')
const cheerio = require('cheerio');

const expect = chai.expect;

/* eslint no-unused-expressions: 0 */
describe('Testing alert notification email templates', async () => {
    const notificationTypes = [
        { type: 'WATCHLIST', title: 'PSC Alert' },
        { type: 'TTP', title: 'PSC Alert' },
        { type: 'POLICY_ACTION', title: 'Policy Action Taken' },
    ];

    notificationTypes.forEach(notification => it(`Should render html correctly for type ${notification.type}`, () => {
        const { html, subject } = formatEmail('pscAlert', Object.assign(
            {},
            context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                type: notification.type,
                title: notification.title
            }
        ));
        const $ = cheerio.load(html); // Load the html so it can be parsed

        // Grab values from html and verify they match the notification data sent in
        const user = $('td:contains("Device username")').next().text();
        const targetValue = $('td:contains("Target value")').next().text();
        const sha256 = $('td:contains("SHA-256")').next().text();
        const incidentId = $('td:contains("Alert ID")').next().text();
        const threatScore = $('td:contains("Severity")').next().text();
        const threatIndicator = $('th:contains("Application")').next().text();

        expect(subject).to.equal(`${notification.title} - ${context.INCIDENT_ID}`);
        expect(user).to.equal(context.DEVICE_EMAIL);
        expect(targetValue).to.equal(context.TARGET_VALUE);
        expect(sha256).to.equal(context.SHA256_HASH);
        expect(incidentId).to.equal(context.INCIDENT_ID);
        expect(threatScore).to.equal(String(context.THREAT_SCORE));
        if (notification.type === 'WATCHLIST') {
            expect(threatIndicator).to.contain('IOCs');
        } else {
            expect(threatIndicator).to.contain('TTPs');
        }

        // Iterate through notification THREAT_INDICATORS,
        // verify their applicationName and corresponding TTPS/IOCs are in the html
        Object.keys(context.THREAT_INDICATORS).forEach((threat) => {
            const appName = context.THREAT_INDICATORS[threat][0].applicationName;
            const application = $(`td:contains(${appName})`);
            expect(application.length, `Expected to find ${appName} in the list of applications.`).to.not.equal(0);
            const indicators = application.next().text();
            context.THREAT_INDICATORS[threat].forEach((indicator) => {
                const message = `Expected to find ${indicator.indicatorName} \
                in the list of TTPs for application ${appName}. Actual: ${indicators}`;
                expect(indicators, message).to.contain(indicator.indicatorName);
            });
        });
    }));
});

describe('Testing CR alert notification email templates', async () => {

    it(`Should render html correctly for CR alerts`, () => {
        const context = Object.assign(
            {},
            crAlertContext.context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                title: 'Carbon Black Cloud Alert',
                links: [
                    {
                        url: `${crAlertContext.context.EXPLICIT_LINKS[crAlertContext.context.PRIMARY_LINK_KEY]}`,
                        label: 'View in Carbon Black Cloud'
                    }
                ]
            });

        const { html, text: fullText, subject } = formatEmail('pscAlert', context);

        expect(subject).to.equal(`${context.title} - ${context.INCIDENT_ID}`);
        expect(html).to.not.be.null;

        const $ = cheerio.load(html);
        const text = fullText.split("\n");

        expectMissingValue($, text, 'User');
        expectPresentValue($, text, 'Target value', context.TARGET_VALUE);
        expectMissingValue($, text, 'Device')
        expectPresentValue($, text, 'Remote host', `${context.REMOTE_DOMAIN} (${context.REMOTE_IP})`);
        expectPresentValue($, text, 'Remote host reputation', context.IP_REPUTATION.toString());
        expectPresentValue($, text, 'Port', context.PORT.toString());
        expectPresentValue($, text, 'Protocol', context.PROTOCOL);
        expectPresentValue($, text, 'Workload', context.WORKLOAD_NAME);
        expectPresentValue($, text, 'Workload kind', context.WORKLOAD_KIND);
        expectPresentValue($, text, 'Namespace', context.NAMESPACE);
        expectPresentValue($, text, 'Cluster', context.CLUSTER_NAME);
        expectMissingValue($, text, 'SHA-256');
        expectMissingValue($, text, 'Process path');
        expectMissingValue($, text, 'Reputation');
        expectPresentValue($, text, 'Alert ID', context.INCIDENT_ID);
        expectPresentValue($, text, 'Severity', context.THREAT_SCORE.toString());
        expectMissingValue($, text, 'Vendor');
        expectMissingValue($, text, 'Vendor ID');
        expectMissingValue($, text, 'Product');
        expectMissingValue($, text, 'Product ID');
        expectMissingValue($, text, 'Watchlist');

        const link = $('a:contains("View in Carbon Black Cloud")');
        expect(link.attr('href')).to.be.equal(context.links[0].url);
    });
});


describe('Testing CR alert HBFW notification email templates', async () => {

    it(`Should render html correctly for CR alerts`, () => {
        const context = Object.assign(
            {},
            crAlertHbfwContext.context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                title: 'Carbon Black Cloud Alert',
                links: [
                    {
                        url: `${crAlertHbfwContext.context.EXPLICIT_LINKS[crAlertHbfwContext.context.PRIMARY_LINK_KEY]}`,
                        label: 'View in Carbon Black Cloud'
                    }
                ]
            });

        const { html, text: fullText, subject } = formatEmail('pscAlert', context);

        expect(subject).to.equal(`${context.title} - ${context.INCIDENT_ID}`);
        expect(html).to.not.be.null;

        const $ = cheerio.load(html);
        const text = fullText.split("\n");


        expectMissingValue($, text, 'Port');
        expectMissingValue($, text, 'Protocol');
        expectMissingValue($, text, 'Workload');
        expectMissingValue($, text, 'Workload kind');
        expectMissingValue($, text, 'Namespace');
        expectMissingValue($, text, 'Cluster');
        expectMissingValue($, text, 'SHA-256');
        expectMissingValue($, text, 'Process path');
        expectMissingValue($, text, 'Reputation');
        expectPresentValue($, text, 'Alert ID', context.INCIDENT_ID);
        expectPresentValue($, text, 'Severity', context.THREAT_SCORE.toString());
        expectMissingValue($, text, 'Vendor');
        expectMissingValue($, text, 'Vendor ID');
        expectMissingValue($, text, 'Product');
        expectMissingValue($, text, 'Product ID');
        expectMissingValue($, text, 'DRE policy ID');
        expectMissingValue($, text, 'Watchlist');

        const link = $('a:contains("View in Carbon Black Cloud")');
        expect(link.attr('href')).to.be.equal(context.links[0].url);
    });
});

describe('Testing CR alert IDS/NTA notification email templates', async () => {

    it(`Should render html correctly for CR alerts`, () => {
        const context = Object.assign(
            {},
            crAlertIdsNtaContext.context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                title: 'Carbon Black Cloud Alert',
                links: [
                    {
                        url: `${crAlertIdsNtaContext.context.EXPLICIT_LINKS[crAlertIdsNtaContext.context.PRIMARY_LINK_KEY]}`,
                        label: 'View in Carbon Black Cloud'
                    }
                ]
            });

        const { html, text: fullText, subject } = formatEmail('pscAlert', context);

        expect(subject).to.equal(`${context.title} - ${context.INCIDENT_ID}`);
        expect(html).to.not.be.null;

        const $ = cheerio.load(html);
        const text = fullText.split("\n");


        expectMissingValue($, text, 'Port');
        expectMissingValue($, text, 'Protocol');
        expectMissingValue($, text, 'Workload');
        expectMissingValue($, text, 'Workload kind');
        expectMissingValue($, text, 'Namespace');
        expectMissingValue($, text, 'Cluster');
        expectMissingValue($, text, 'SHA-256');
        expectMissingValue($, text, 'Process path');
        expectMissingValue($, text, 'Reputation');
        expectPresentValue($, text, 'Alert ID', context.INCIDENT_ID);
        expectPresentValue($, text, 'Severity', context.THREAT_SCORE.toString());
        expectMissingValue($, text, 'Vendor');
        expectMissingValue($, text, 'Vendor ID');
        expectMissingValue($, text, 'Product');
        expectMissingValue($, text, 'Product ID');
        expectMissingValue($, text, 'DRE policy ID');
        expectPresentValue($, text, 'TMS Rule ID', context.TMS_RULE_ID);
        expectMissingValue($, text, 'Watchlist');

        const link = $('a:contains("View in Carbon Black Cloud")');
        expect(link.attr('href')).to.be.equal(context.links[0].url);
    });
});

describe('Testing CR alert MDR Analyst added comment notification email templates', async () => {

    it(`Should render html correctly for CR alerts MDR Analyst added comment`, () => {
        const context = Object.assign(
            {},
            crAlertMDRCommentAdded.context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                title: 'Managed Detection and Response Comment Added',
                links: [
                    {
                        url: `${crAlertMDRCommentAdded.context.EXPLICIT_LINKS[crAlertMDRCommentAdded.context.PRIMARY_LINK_KEY]}`,
                        label: 'Log in to View Comment'
                    }
                ]
            });

        const { html, text: fullText, subject } = formatEmail('mdrCommentAdded', context);
        const num = 5;
        expect(subject).to.equal(`${context.title} - ${context.INCIDENT_ID}`);
        expect(html).to.not.be.null;

        const $ = cheerio.load(html);
        const text = fullText.split("\n");

        expectPresentValue($, text, 'Alert ID', context.INCIDENT_ID);
        expectPresentValue($, text, 'Alert severity', context.THREAT_SCORE.toString());
        expectPresentValue($, text, 'Alert type', context.MDR_ALERT_TYPE);
        expectPresentValue($, text, 'Device name', context.DEVICE_NAME);
        expectPresentValue($, text, 'Device ID', context.DEVICE_ID.toString());
        expectPresentValue($, text, 'Target value', context.TARGET_VALUE);
        expectPresentValue($, text, 'Process username', context.PROCESS_USER_NAME);
        expectPresentValue($, text, 'MDR workflow', 'In progress');

        const link = $('a:contains("Log in to View Comment")');
        expect(link.attr('href')).to.be.equal(context.links[0].url);
    });
});

describe('Testing CR alert MDR Likely Threat notification email templates', async () => {

    it(`Should render html correctly for CR alerts MDR Likely Threat`, () => {
        const context = Object.assign(
            {},
            crAlertMDRLikelyThreat.context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                title: 'Managed Detection and Response - Likely Threat',
                links: [
                    {
                        url: `${crAlertMDRLikelyThreat.context.EXPLICIT_LINKS[crAlertMDRLikelyThreat.context.PRIMARY_LINK_KEY]}`,
                        label: 'Log in to View Alert'
                    }
                ]
            });

        const { html, text: fullText, subject } = formatEmail('mdrLikelyThreat', context);

        expect(subject).to.equal(`${context.title} - ${context.INCIDENT_ID}`);
        expect(html).to.not.be.null;

        const $ = cheerio.load(html);
        const text = fullText.split("\n");

        expectPresentValue($, text, 'Alert ID', context.INCIDENT_ID);
        expectPresentValue($, text, 'Alert severity', context.THREAT_SCORE.toString());
        expectPresentValue($, text, 'Alert type', context.MDR_ALERT_TYPE);
        expectPresentValue($, text, 'Device name', context.DEVICE_NAME);
        expectPresentValue($, text, 'Device ID', context.DEVICE_ID.toString());
        expectPresentValue($, text, 'Target value', context.TARGET_VALUE);
        expectPresentValue($, text, 'Process username', context.PROCESS_USER_NAME);
        expectPresentValue($, text, 'MDR workflow', 'Triage complete');
        expectPresentValue($, text, 'MDR determination', 'Likely threat');

        const link = $('a:contains("Log in to View Alert")');
        expect(link.attr('href')).to.be.equal(context.links[0].url);
    });
});

describe('Testing CR alert MDR Unlikely Threat notification email templates', async () => {

    it(`Should render html correctly for CR alerts MDR Unlikely Threat`, () => {
        const context = Object.assign(
            {},
            crAlertMDRUnlikelyThreat.context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                title: 'Managed Detection and Response - Unlikely Threat',
                links: [
                    {
                        url: `${crAlertMDRUnlikelyThreat.context.EXPLICIT_LINKS[crAlertMDRUnlikelyThreat.context.PRIMARY_LINK_KEY]}`,
                        label: 'Log in to View Alert'
                    }
                ]
            });

        const { html, text: fullText, subject } = formatEmail('mdrUnlikelyThreat', context);

        expect(subject).to.equal(`${context.title} - ${context.INCIDENT_ID}`);
        expect(html).to.not.be.null;

        const $ = cheerio.load(html);
        const text = fullText.split("\n");

        expectPresentValue($, text, 'Alert ID', context.INCIDENT_ID);
        expectPresentValue($, text, 'Alert severity', context.THREAT_SCORE.toString());
        expectPresentValue($, text, 'Alert type', context.MDR_ALERT_TYPE);
        expectPresentValue($, text, 'Device name', context.DEVICE_NAME);
        expectPresentValue($, text, 'Device ID', context.DEVICE_ID.toString());
        expectPresentValue($, text, 'Target value', context.TARGET_VALUE);
        expectPresentValue($, text, 'Process username', context.PROCESS_USER_NAME);
        expectPresentValue($, text, 'MDR workflow', 'Triage complete');
        expectPresentValue($, text, 'MDR determination', 'Unlikely threat');

        const link = $('a:contains("Log in to View Alert")');
        expect(link.attr('href')).to.be.equal(context.links[0].url);
    });
});

describe('Testing CR alert MDR Threat Hunt likely threat found notification email templates', async () => {

    it(`Should render html correctly for CR alerts MDR Unlikely Threat`, () => {
        const context = Object.assign(
            {},
            crAlertMDRThreatHuntLikelyThreatFound.context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                title: 'Managed Detection and Response Threat Hunt - Likely Threat Found',
                threatHuntInvestigatePageLinks: [
                    {
                        url: `${crAlertMDRThreatHuntLikelyThreatFound.context.THREAT_HUNT_INVESTIGATE_PAGE_LINK}`,
                        label: 'Log In to View Hits'
                    }
                ],
                threatHuntAlertsPageLinks: [
                    {
                        url: `${crAlertMDRThreatHuntLikelyThreatFound.context.THREAT_HUNT_ALERT_PAGE_LINK}`,
                        label: 'Log In to View Likely Threat Alerts'
                    }
                ]
            });

        const { html, text: fullText, subject } = formatEmail('mdrThreatHuntThreatFound', context);

        expect(subject).to.equal(`${context.title}`);
        expect(html).to.not.be.null;

        const $ = cheerio.load(html);
        const text = fullText.split("\n");

        expectPresentValue($, text, 'Threat hunt name', context.THREAT_HUNT_NAME);
        expectPresentValue($, text, 'Threat hunt completed', context.THREAT_HUNT_CLOSED_TIMESTAMP);
        expectPresentValue($, text, 'Threat hunt report', context.THREAT_HUNT_REPORT_NAME);
        expectPresentValue($, text, 'Threat hunt description', context.THREAT_HUNT_DESCRIPTION);

        const link1 = $('a:contains("Log In to View Hits")');
        expect(link1.attr('href')).to.be.equal(context.threatHuntInvestigatePageLinks[0].url);
    });
});

describe('Testing CR alert MDR Threat Hunt No threats found notification email templates', async () => {

    it(`Should render html correctly for CR alerts MDR Unlikely Threat`, () => {
        const context = Object.assign(
            {},
            crAlertMDRThreatNoThreatsFound.context,
            {
                staticHost: 'https://defense-dev01.cbdtest.io',
                title: 'Managed Detection and Response Threat Hunt - No Threats Found',
                threatHuntInvestigatePageLinks: [
                    {
                        url: `${crAlertMDRThreatNoThreatsFound.context.THREAT_HUNT_INVESTIGATE_PAGE_LINK}`,
                        label: 'Log In to View Hits'
                    }
                ]
            });

        const { html, text: fullText, subject } = formatEmail('mdrThreatHuntNoThreatFound', context);

        expect(subject).to.equal(`${context.title}`);
        expect(html).to.not.be.null;

        const $ = cheerio.load(html);
        const text = fullText.split("\n");

        expectPresentValue($, text, 'Threat hunt name', context.THREAT_HUNT_NAME);
        expectPresentValue($, text, 'Threat hunt completed', context.THREAT_HUNT_CLOSED_TIMESTAMP);
        expectPresentValue($, text, 'Threat hunt report', context.THREAT_HUNT_REPORT_NAME);
        expectPresentValue($, text, 'Threat hunt description', context.THREAT_HUNT_DESCRIPTION);

        const link1 = $('a:contains("Log In to View Hits")');
        expect(link1.attr('href')).to.be.equal(context.threatHuntInvestigatePageLinks[0].url);
    });
});

const expectPresentValue = (html, text, titleText, expected) => {
    const titleCell = getTableCellByExactContent(html, titleText);
    expect(titleCell).to.have.lengthOf.above(0);
    const value = titleCell.next().text();
    expect(value).to.equal(expected);
    expect(text).to.contain(`    ${titleText}: ${expected}`);
}

const expectMissingValue = (html, text, titleText) => {
    const titleCell = getTableCellByExactContent(html, titleText);
    expect(titleCell).to.have.lengthOf(0);
    expect(text).to.not.contain(`    ${titleText}: `);
}

const getTableCellByExactContent = ($, content) => {
    return $('td').filter(function() {
        return $(this).text().trim() === content;
    });
}
