const itParam = require('mocha-param');
const request = require('supertest');
const chai = require('chai');
const Users = require('../models/users');
const Notifications = require('../models/notifications');
const config = require('../config');
const { createUser } = require('./helpers/createUser');

const expect = chai.expect;
const server = `${config.HOST}:${config.PORT}/notifications/v1`;

/* eslint no-unused-expressions: 0 */
describe('Testing notifications', () => {
    let user;
    let user2;
    before(async () => {
        user = await createUser('test_env', 1, 1, 1, 1);
        user2 = await createUser('test_env2', 1, 1, 1, 1);
    });

    after(async () => {
        await Users.sync({ force: true });
        await Notifications.sync({ force: true });
    });

    describe('Testing getting notifications', () => {
        it('Should get notifications successfully', async () => {
            const response = await request(server).get('/notifications').set(user.headers);
            expect(response.status).to.equal(200);
            expect(response.body).to.have.property('unReadCount');
            expect(response.body).to.have.property('items');
        });
    });

    describe('Testing posting a notification with invalid URL', async () => {
        const urlData = ['https://notcarbonblack.com', 'https://evil.com?foo=https://carbonblack.com',
            'https://carbonblack.com.evil.com', 'https://evil.com/carbonblack.com',
            'https://carbonblack/com/evil.com'];
        itParam('Should fail to post notifications linking to a non CB site', urlData, async (url) => {
            const data = {
                title: 'Test',
                content: 'Testing invalid URL',
                action: {
                    url
                }
            };
            const response = await request(server).post('/notifications').set(user.headers).send(data);
            expect(response.status).to.not.equal(200);
            expect(response.body.error).to.equal('Only Cb Domains can currently be linked to.');
        });
    });

    describe('Testing posting a notification with valid URL', async () => {
        const data = {
            title: 'Test notification',
            content: 'Testing valid url',
            action: {
                url: 'https://carbonblack.com'
            }
        };
        it('Should post a notification using a valid URL to all environments', async () => {
            const response = await request(server).post('/notifications').set(user.headers).send(data);
            expect(response.status).to.equal(200);
            expect(response.body.message).to.equal('Operation succeeded.');
            expect(response.body.count).to.be.at.least(1);
        });

        it('Should show posted notification in list of notifications', async () => {
            // Verify that the most recent notification returned is the one just posted
            const response = await request(server).get('/notifications').set(user.headers);
            expect(response.status).to.equal(200);
            expect(response.body).to.have.property('items');
            expect(response.body.items).to.not.be.empty;
            expect(response.body.items[0].title).to.equal(data.title);
            expect(response.body.items[0].content).to.equal(data.content);
            expect(response.body.items[0].action).to.deep.equal(data.action);
        });

        it('Should show notification for user in different environment', async () => {
            const response = await request(server).get('/notifications').set(user2.headers);
            expect(response.status).to.equal(200);
            expect(response.body.items).to.not.be.empty;
            expect(response.body.items[0].title).to.equal(data.title);
            expect(response.body.items[0].content).to.equal(data.content);
            expect(response.body.items[0].action).to.deep.equal(data.action);
        });
    });

    describe('Testing posting a notification with valid Broadcom URL', async () => {
        const data = {
            title: 'Test new URL notification',
            content: 'Testing valid Broadcom url',
            action: {
                url: 'https://community.broadcom.com'
            }
        };
        it('Should post a notification using a valid URL to all environments', async () => {
            const response = await request(server).post('/notifications').set(user.headers).send(data);
            expect(response.status).to.equal(200);
            expect(response.body.message).to.equal('Operation succeeded.');
            expect(response.body.count).to.be.at.least(1);
        });

        it('Should show posted notification in list of notifications', async () => {
            // Verify that the most recent notification returned is the one just posted
            const response = await request(server).get('/notifications').set(user.headers);
            expect(response.status).to.equal(200);
            expect(response.body).to.have.property('items');
            expect(response.body.items).to.not.be.empty;
            expect(response.body.items[0].title).to.equal(data.title);
            expect(response.body.items[0].content).to.equal(data.content);
            expect(response.body.items[0].action).to.deep.equal(data.action);
        });

        it('Should show notification for user in different environment', async () => {
            const response = await request(server).get('/notifications').set(user2.headers);
            expect(response.status).to.equal(200);
            expect(response.body.items).to.not.be.empty;
            expect(response.body.items[0].title).to.equal(data.title);
            expect(response.body.items[0].content).to.equal(data.content);
            expect(response.body.items[0].action).to.deep.equal(data.action);
        });
    });

    describe('Testing dismissing a notification', async () => {
        let itemsCount;
        let notificationId;
        it('Should POST a notification', async () => {
            const response = await request(server)
                .post('/notifications')
                .set(user.headers)
                .send({ title: 'Test', content: 'Notification to be dismissed' });
            expect(response.status).to.equal(200);

            // Get notificationId and item count for recently added notifications
            const notifications = await request(server).get('/notifications').set(user.headers);
            expect(notifications.body.items[0]).to.have.property('id');
            notificationId = notifications.body.items[0].id;
            itemsCount = notifications.body.items.length;
        });

        it('Should dismiss notification', async () => {
            // Set notification as dismissed, using id of most recently added notification
            const response = await request(server)
                .patch(`/notifications/${notificationId}`)
                .set(user.headers)
                .send({ dismissed: true });
            expect(response.status).to.equal(200);
            expect(response.body.update).to.deep.equal([1]);
        });

        it('Should not show dismissed notifications', async () => {
            // The item count should go down by one and the notification
            // should not exist in the list of notifications
            const response = await request(server)
                .get('/notifications')
                .set(user.headers);
            expect(response.status).to.equal(200);
            expect(response.body).to.have.property('items');
            expect(response.body.items.length).to.equal(itemsCount - 1);
            const notification = response.body.items.find(item => item.id === notificationId);
            expect(notification).to.be.undefined;
        });
    });

    describe('Testing reading a notification', async () => {
        let notificationId;
        const readTime = new Date().toISOString();
        it('Should POST a notification', async () => {
            const response = await request(server)
                .post('/notifications')
                .set(user.headers)
                .send({ title: 'Test', content: 'Notification to be read' });
            expect(response.status).to.equal(200);
        });

        it('Should read notification', async () => {
            const response = await request(server)
                .get('/notifications')
                .set(user.headers);
            expect(response.body.items[0]).to.have.property('id');
            notificationId = response.body.items[0].id;

            // Set notification as read, using id of most recently added notification
            const readResponse = await request(server)
                .patch(`/notifications/${notificationId}`)
                .set(user.headers)
                .send({ readtime: readTime });
            expect(readResponse.status).to.equal(200);
            expect(readResponse.body.update).to.deep.equal([1]);
        });

        it('Should show notification as read', async () => {
            const response = await request(server)
                .get('/notifications')
                .set(user.headers);
            expect(response.status).to.equal(200);
            expect(response.body).to.have.property('items');
            expect(response.body.items).to.not.be.empty;
            const notification = response.body.items.find(item => item.id === notificationId);
            expect(notification).to.not.be.undefined;
            expect(notification.readtime).to.equal(readTime);
        });
    });

    describe('Testing getting notifications reach', async () => {
        it('Should reach at least one user on all environments', async () => {
            const response = await request(server).post('/reach').set(user.headers);
            expect(response.status).to.equal(200);
            expect(response.body.count).to.be.at.least(1);
        });

        it('Should return 0 for an environment with no users', async () => {
            const response = await request(server)
                .post('/reach')
                .set(user.headers)
                .send({ where: { environment: 'bogusEnvironment' } });
            expect(response.status).to.equal(200);
            expect(response.body.count).to.equal(0);
        });
    });

    describe('Testing posting notification to individual environment', async () => {
        let notificationId;
        it('Should post notification to individual environment', async () => {
            const data = {
                title: 'Test notification',
                content: 'Notification for test env only',
                type: 'announcement',
                where: { environment: 'test_env' }
            };
            const response = await request(server)
                .post('/notifications')
                .set(user.headers)
                .send(data);
            expect(response.status).to.equal(200);
            const notifications = await request(server).get('/notifications').set(user.headers);
            expect(notifications.status).to.equal(200);
            expect(notifications.body.items[0].content).to.equal(data.content);
            expect(notifications.body.items[0]).to.have.property('id');
            notificationId = notifications.body.items[0].id;
        });

        it('Should not show notification in different environment', async () => {
            // Get notifications as user in a different environment
            const response = await request(server).get('/notifications').set(user2.headers);
            expect(response.status).to.equal(200);
            if (response.body.items.length !== 0) {
                const notification = response.body.items.find(item => item.id === notificationId);
                expect(notification).to.be.undefined;
            }
        });
    });
});
