const handlebars = require("handlebars");
const fs = require("fs");
const path = require("path");

handlebars.registerHelper('equal', (lvalue, rvalue, options) => {
    if (arguments.length < 3) {
        throw new Error('Handlebars Helper equal needs 2 parameters');
    } else if (lvalue !== rvalue) {
        return options.inverse(this);
    } else {
        return options.fn(this);
    }
});

// baseTemplate is the common wrapper for all HTML emails. Each email is rendered in two steps:
//   1. Render the partial for the specific email being sent
//   2. Render the base template, with the output from the first render filled as the {{ content }}
const baseTemplate = handlebars.compile(
    String(fs.readFileSync(path.join(__dirname, '../../templates', 'base_template.html')))
);

const mdrBaseTemplate = handlebars.compile(
    String(fs.readFileSync(path.join(__dirname, '../../templates', 'mdr_base_template.html')))
);

// createTemplate loads the template file from disk and fills the renderHtml and renderText functions
// to actually generate the email content
const createTemplate = (name, title) => {
    const template = {
        title
    };

    try {
        const renderHtml = handlebars.compile(
            String(fs.readFileSync(path.join(__dirname, '../../templates/partials', `${name}.html`)))
        )

        template.renderHtml = data => baseTemplate({
            ...data,
            title,
            content: renderHtml(data)
        });
    } catch (err) {
        console.error(err.message);
        throw new Error(`HTML template partial not found ${name}`);
    }


    try {
        template.renderText = handlebars.compile(
            String(fs.readFileSync(path.join(__dirname, '../../templates/partials', `${name}.txt`)))
        );
    } catch (err) {
        // Text fallback not strictly enforced.
    }

    return template;
}

const mdrCreateTemplate = (name, topTitle, bottomTitle) => {
    const template = {
        topTitle,
        bottomTitle
    };

    try {
        const renderHtml = handlebars.compile(
            String(fs.readFileSync(path.join(__dirname, '../../templates/partials', `${name}.html`)))
        )

        template.renderHtml = data => mdrBaseTemplate({
            ...data,
            topTitle,
            bottomTitle,
            content: renderHtml(data)
        });
    } catch (err) {
        console.error(err.message);
        throw new Error(`HTML template partial not found ${name}`);
    }

    try {
        template.renderText = handlebars.compile(
            String(fs.readFileSync(path.join(__dirname, '../../templates/partials', `${name}.txt`)))
        );
    } catch (err) {
        // Text fallback not strictly enforced.
    }

    return template;
}

// Known email templates
const templates = {
    mdrCommentAdded: mdrCreateTemplate('mdrCommentAdded', 'Managed Detection & Response', 'Comment Added'),
    mdrCommentAddedForCbAnalytics: mdrCreateTemplate('mdrCommentAddedForCbAnalytics', 'Managed Detection & Response', 'Comment Added'),
    mdrCommentAddedForWatchlist: mdrCreateTemplate('mdrCommentAddedForWatchlist', 'Managed Detection & Response', 'Comment Added'),
    mthCommentAddedForWatchlist: mdrCreateTemplate('mdrCommentAddedForWatchlist', 'Managed Threat Hunting', 'Comment Added'),
    mdrLikelyThreat: mdrCreateTemplate('mdrLikelyOrUnlikelyThreat', 'Managed Detection & Response', 'Likely Threat'),
    mdrLikelyThreatForCbAnalytics: mdrCreateTemplate('mdrLikelyOrUnlikelyThreatForCbAnalytics', 'Managed Detection & Response', 'Likely Threat'),
    mdrLikelyThreatForWatchlist: mdrCreateTemplate('mdrLikelyOrUnlikelyThreatForWatchlist', 'Managed Detection & Response', 'Likely Threat'),
    mthLikelyThreatForWatchlist: mdrCreateTemplate('mdrLikelyOrUnlikelyThreatForWatchlist', 'Managed Threat Hunting', 'Likely Threat'),
    mdrUnlikelyThreat: mdrCreateTemplate('mdrLikelyOrUnlikelyThreat', 'Managed Detection & Response', 'Unlikely Threat'),
    mdrUnlikelyThreatForCbAnalytics: mdrCreateTemplate('mdrLikelyOrUnlikelyThreatForCbAnalytics', 'Managed Detection & Response', 'Unlikely Threat'),
    mdrUnlikelyThreatForWatchlist: mdrCreateTemplate('mdrLikelyOrUnlikelyThreatForWatchlist', 'Managed Detection & Response', 'Unlikely Threat'),
    mthUnlikelyThreatForWatchlist: mdrCreateTemplate('mdrLikelyOrUnlikelyThreatForWatchlist', 'Managed Threat Hunting', 'Unlikely Threat'),
    mdCommentAddedForCbAnalytics: mdrCreateTemplate('mdrCommentAddedForCbAnalytics', 'Managed Detection', 'Comment Added'),
    mdLikelyThreatForCbAnalytics: mdrCreateTemplate('mdrLikelyOrUnlikelyThreatForCbAnalytics', 'Managed Detection', 'Likely Threat'),
    mdUnlikelyThreatForCbAnalytics: mdrCreateTemplate('mdrLikelyOrUnlikelyThreatForCbAnalytics', 'Managed Detection', 'Unlikely Threat'),
    mdrThreatHuntThreatFound: mdrCreateTemplate('mdrThreatHunt', 'Managed Threat Hunting', 'Likely Threats Found'),
    mdrThreatHuntNoThreatFound: mdrCreateTemplate('mdrThreatHunt', 'Managed Threat Hunting', 'No Threats Found'),
    pscAlert: createTemplate('pscAlert', 'Carbon Black Cloud Alert'),
    pscMDREligibleCBAnalyticsAlert: createTemplate('pscMDREligibleCBAnalyticsAlert', 'Carbon Black Cloud Alert'),
    pscMDREligibleWatchlistAlert: createTemplate('pscMDREligibleWatchlistAlert', 'Carbon Black Cloud Alert'),
    policyActionApplied: createTemplate('pscAlert', 'Carbon Black Cloud Action'),
    emailQuotaExceeded: createTemplate('emailQuotaExceeded', 'Email notification quota exceeded'),
    test: createTemplate('test', 'Carbon Black - email test delivery'),
};

module.exports = { templates }
