const compose = require('koa-compose');
const json = require('koa-json');
const bodyParser = require('koa-body');
const requestIp = require('request-ip');
const config = require('../config');
const Users = require('../models/users');
const responses = require('../utils/responses');
const checkPermission = require('../utils/checkPermission');
const saveAuditLog = require('../utils/auditLog');

const gatewayurl = config.GATEWAY_URL;
const shouldForwardHostHeader = config.GATEWAY_FORWARD_HOST === 'true';

const IGNORE_URLS = [
    `/${config.SERVICE_NAME}/${config.VERSION}/health_check`,
    `/${config.SERVICE_NAME}/${config.VERSION}/notifications`,
];

const logRequest = () => async (ctx, next) => {
    const start = Date.now();
    ctx.set('Access-Control-Allow-Origin', '*');
    ctx.set('Access-Control-Allow-Headers', 'Content-Type, Access-Control-Allow-Headers, Authorization, X-Requested-With, X-CSRF-Token, X-Forward-Token, Preferred-Locale');
    ctx.set('Access-Control-Allow-Methods', 'GET, POST, DELETE, PATCH, PUT, OPTIONS');
    ctx.body = {};
    ctx.status = 404;
    // Limit fields to load to exposed fields
    const attributes = Users.fields.filter(f => f.exposed).map(f => f.name);
    try {
        if (ctx.request.headers['x-user-context']) {
            ctx.state.gateway_user = JSON.parse(ctx.request.headers['x-user-context']);
        }
        if (ctx.request.headers['x-request-context']) {
            ctx.state.gateway = JSON.parse(ctx.request.headers['x-request-context']);
        }
        if (ctx.state.gateway && ctx.state.gateway_user) {
            const where = {
                loginid: ctx.state.gateway_user.target_org.user_id,
                environment: ctx.state.gateway.origin_env,
                orgid: ctx.state.gateway_user.target_org.org_id
            };
            // Attempt finding the user sent down from gateway
            ctx.state.user = await Users.findOne({ where, attributes });
            if (!ctx.state.user) {
                // User not found in the service, set him up.
                ctx.state.user = await Users.create({
                    environment: ctx.state.gateway.origin_env,
                    orgid: ctx.state.gateway_user.target_org.org_id,
                    orgkey: ctx.state.gateway_user.target_org.org_key,
                    loginid: ctx.state.gateway_user.target_org.user_id,
                    orgrole: ctx.state.gateway_user.target_org.rbac_key
                });
            }
        }
    } catch (error) {
        ctx.body.authError = error.message;
        console.log(`Auth failed, error: ${error.message}`)
    }

    ctx.state.checkPermission = (permission, action) => checkPermission(
        permission,
        action,
        ctx.state.gateway,
        ctx.state.gateway_user,
        ctx.request.headers
    );

    ctx.state.auditLog = description => ctx.state.gateway_user && saveAuditLog(
        {
            description,
            clientIp: requestIp.getClientIp(ctx.request),
            requestUrl: ctx.request.url,
            orgId: ctx.state.gateway_user.login_org.org_id,
            loginName: ctx.state.gateway_user.login_org.user_name,
            eventTime: Date.now()
        },
        ctx.state.gateway,
        ctx.state.gateway_user,
        ctx.request.headers,
        gatewayurl,
        shouldForwardHostHeader
    );

    try {
        await next();
    } catch (err) {
        ctx.status = err.status || 500;
        ctx.body.error = err.message || 'Server error';
    }
    if (!ctx.body.status) {
        ctx.body.status = responses[ctx.status] && responses[ctx.status].description;
    }
    const execution_time = Date.now() - start;

    ctx.set('X-Application-Execution-Time', `${execution_time}ms`);
    ctx.set('X-Application-Hash', config.HASH);

    // Log all non 200 requests, or request that aren't in the ignore list
    // Ignore /health_check and /notifications since those are pulled programmatically.
    if (IGNORE_URLS.every(url => !ctx.url.startsWith(url)) || ctx.status !== 200 || ctx.method !== 'GET') {
        console.log(`${ctx.status} ${ctx.method} ${ctx.url} ${execution_time}ms: ${ctx.body.status} ${ctx.body.error ? ctx.body.error : ''}`);
    }
};

module.exports = compose([
    bodyParser({ multipart: true, maxFieldsSize: 30 * 1024 * 1024 }),
    json({ pretty: config.PRETTY_JSON }),
    logRequest()
]);
